module System.Classes;

import delta.core;
import std.utf: toUTFz;

alias TNotifyEvent = void delegate(TObject sender);

void executeNotifyEvent(TObject object, string delphiEventName, void* framePointer, void* fnPointer)
{
	import core.sys.windows.windows;
	alias extern(Windows) void function(int, int) NotifyEventCallback;
	alias extern(Windows) void function(int, char*, int, NotifyEventCallback) Fn;
	
	void delegate(int, int) dg = (int delphiRef, int dRef) {
		TNotifyEvent ne;
		ne.funcptr = cast(void function(TObject)) fnPointer;
		ne.ptr = cast(void*) dRef;
		ne(cast(TObject) cast(void*) dRef);
	};
	
	auto pChar = toUTFz!(char*)(delphiEventName);
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "registerNotifyEvent");
	Fn fn = cast(Fn) fp;
	fn(object.reference, pChar, cast(int) framePointer, bindDelegate(dg));
}

enum TShiftState {ssShift, ssAlt, ssCtrl, ssLeft, ssRight, ssMidldle, ssDouble, ssTouch, ssPen, ssCommand, ssHorizontal};

class TBasicAction: TComponent
{
	mixin(GenNotifyEvent!("OnExecute"));
	
	this(int reference)
	{
		super(reference);
	}
}

class TComponent: TPersistent
{
	T opCast(T)() {
        return new T(reference);
    }
	
	TComponent FindComponent(string AName)
	{
		int i = executeInstanceMethodReturnReferenceArgsString(this.reference, "FindComponent", AName);
		return (i == 0) ? null : new TComponent(i);
	}

	this(int reference)
	{
		super(reference);
	}
}

class TStrings: TPersistent
{
	mixin DelphiClass!("System.Classes.TStrings");
	@property string Text();
	@property void Text(string value);
	
	int Add(string value);
	
	int AddObject(string value, TObject aObject);
	
	void Clear();
}

class TStringList: TStrings
{
	mixin DelphiClass!("System.Classes.TStringList");
}

class TStream : TObject
{
	mixin DelphiClass!("System.Classes.TStream");
}

class TPersistent: TObject
{
	mixin DelphiClass!("System.Classes.TPersistent");

	void Destroy();
}

class TInterfacedPersistent: TPersistent
{
	mixin DelphiClass!("System.Classes.TInterfacedPersistent");
}

class TObject
{
	protected int _reference;
	
	@property int reference()
	{
		return _reference;
	}

	T opCast(T)() {
        return new T(reference);
    }
	
	this(int reference)
	{
		_reference = reference;
	}
	
	void DisposeOf()
	{
		executeInstanceMethodReturnNoneArgsNone(this.reference, "DisposeOf");
	}
}

class TCollection: TPersistent
{
	mixin DelphiClass!("System.Classes.TCollection");
	@property int Count();
	
	TCollectionItem Add();

	void Clear();

	string GetNamePath();
}

class TCollectionItem: TPersistent
{
	mixin DelphiClass!("System.Classes.TCollectionItem");
}

class TOwnedCollection : TCollection
{
    mixin DelphiClass!("FMX.Types.TOwnedCollection");

	/*static TOwnedCollection Create(TPersistent AOwner, TCollectionItemClass ItemClass: )
	{
		// int reference = executeClassMethodReturnReferenceArgsNone("System.Classes.TOwnedCollection", "Create");
		return new TOwnedCollection(reference);
	}*/
}

interface IStreamPersist
{
	void LoadFromStream(TStream Stream);
    void SaveToStream(TStream Stream);
}