module delta.core.methods;

import core.sys.windows.windows;
import core.sys.windows.wtypes: BSTR;
import std.utf: toUTFz, encode;
import std.algorithm: each;
import std.conv: to;

import delta.core.dll;

bool isProperty(string[] items...) {
	foreach(item; items)
		if(item == "@property")
			return true;
	return false;
}

enum skip;

mixin template DelphiClass(string name)
{
	private mixin template DelphiImportImpl(T, alias method, size_t overloadIndex)
	{
		import std.traits;

		static if(isProperty(__traits(getFunctionAttributes, method))) {
			static assert(__traits(isStaticFunction, method) == false);

			static if (is(ReturnType!method == void) && Parameters!method.length == 1)
			{
				pragma(mangle, method.mangleof)
				private void implementation(Parameters!method args)
				{
					static if(is(Parameters!method[0] == bool))
					{
						setPropertyBool(_reference, __traits(identifier, method), args[0]);
					}
					else static if(is(Parameters!method[0] == float))
					{
						setPropertyFloat(_reference, __traits(identifier, method), args[0]);
					}
					else static if(is(Parameters!method[0] == short))
					{
						setPropertyShort(_reference, __traits(identifier, method), args[0]);
					}
					else static if(is(Parameters!method[0] == int))
					{
						setPropertyInt(_reference, __traits(identifier, method), args[0]);
					}
					else static if(is(Parameters!method[0] == uint))
					{
						setPropertyUInt(_reference, __traits(identifier, method), args[0]);
					}
					else static if(is(Parameters!method[0] == string))
					{
						setPropertyString(_reference, __traits(identifier, method), args[0]);
					}
					else static assert(false, "Not implemented");
				}
			}
			else static if (!is(ReturnType!method == void) && Parameters!method.length == 0)
			{
				pragma(mangle, method.mangleof)
				private ReturnType!method implementation()
				{
					static if(is(ReturnType!method == bool))
					{
						return getPropertyBool(_reference, __traits(identifier, method));
					}
					else static if(is(ReturnType!method == float))
					{
						return getPropertyFloat(_reference, __traits(identifier, method));
					}
					else static if(is(ReturnType!method == short))
					{
						return getPropertyShort(_reference, __traits(identifier, method));
					}
					else static if(is(ReturnType!method == int))
					{
						return getPropertyInt(_reference, __traits(identifier, method));
					}
					else static if(is(ReturnType!method == uint))
					{
						return getPropertyUInt(_reference, __traits(identifier, method));
					}
					else static if(is(ReturnType!method == string))
					{
						return getPropertyString(_reference, __traits(identifier, method));
					}
					else static assert(false, "Not implemented");
				}
			} else static assert(false, "Not implemented");
		} 
		else 
		{

			static if (__traits(isStaticFunction, method))
			{
				pragma(mangle, method.mangleof)
				private static ReturnType!method implementation(Parameters!method args)
				{
					//	executeClassMethodReturnReferenceArgsReference
					static if(is(ReturnType!method : Object) && args.length == 1 && is(typeof(args[0]) : Object))
					{
						int arg0Reference = args[0] is null ? nil : args[0].reference;
						int resultReference = executeClassMethodReturnReferenceArgsReference(name, __traits(identifier, method), arg0Reference);
						return new ReturnType!method(resultReference);
					} 
					// executeClassMethodReturnReferenceArgsString
					else static if(is(ReturnType!method : Object) && args.length == 1 && is(typeof(args[0]) == string))
					{
						int resultReference = executeClassMethodReturnReferenceArgsString(name, __traits(identifier, method), args[0]);
						return new ReturnType!method(resultReference);
					}
					// executeClassMethodReturnReferenceArgsNone
					else static if(is(ReturnType!method : Object) && args.length == 0)
					{
						int resultReference = executeClassMethodReturnReferenceArgsNone(name, __traits(identifier, method));
						return new ReturnType!method(resultReference);
					}

					else static assert(false, "Not implemented: " ~ __traits(identifier, method));
				}
				/*
				executeClassMethodReturnNoneArgsNone
				executeClassMethodReturnReferenceArgsIntUInt
				executeClassMethodReturnNoneArgsString
				executeClassMethodReturnStringArgsStringStringString
				executeClassMethodReturnReferenceArgsReferenceInt*/
			}
			else
			{
				pragma(mangle, method.mangleof)
				private ReturnType!method implementation(Parameters!method args)
				{
					//	executeInstanceMethodReturnBoolArgsNone
					static if(is(ReturnType!method == bool) && args.length == 0)
					{
						return executeInstanceMethodReturnBoolArgsNone(_reference, __traits(identifier, method));
					}


					// executeInstanceMethodReturnNoneArgsReference
					else static if(is(ReturnType!method == void) && args.length == 1 && is(typeof(args[0]) : Object))
					{
						executeInstanceMethodReturnNoneArgsReference(_reference, __traits(identifier, method), args[0].reference);
					}
					// executeInstanceMethodReturnNoneArgsString
					else static if(is(ReturnType!method == void) && args.length == 1 && is(typeof(args[0]) == string))
					{
						executeInstanceMethodReturnNoneArgsString(_reference, __traits(identifier, method), args[0]);
					}
					// executeInstanceMethodReturnNoneArgsNone
					else static if(is(ReturnType!method == void) && args.length == 0)
					{
						executeInstanceMethodReturnNoneArgsNone(_reference, __traits(identifier, method));
					}
					
					
					// executeInstanceMethodReturnStringArgsNone
					else static if(is(ReturnType!method == string) && args.length == 0)
					{
						return executeInstanceMethodReturnStringArgsNone(_reference, __traits(identifier, method));
					}

					// executeInstanceMethodReturnReferenceArgsNone
					else static if(is(ReturnType!method : Object) && args.length == 0)
					{
						return new TCollectionItem(executeInstanceMethodReturnReferenceArgsNone(_reference, __traits(identifier, method)));
					}

					// executeInstanceMethodReturnIntArgsString
					else static if(is(ReturnType!method == int) && args.length == 1 && is(typeof(args[0]) == string))
					{
						return executeInstanceMethodReturnIntArgsString(_reference, __traits(identifier, method), args[0]);
					}
					// executeInstanceMethodReturnIntArgsStringReference
					else static if(is(ReturnType!method == int) && args.length == 2 && is(typeof(args[0]) == string ) && is(typeof(args[1]) : Object))
					{
						return executeInstanceMethodReturnIntArgsStringReference(_reference, __traits(identifier, method), args[0], args[1].reference);
					}

					else static assert(false, "Not implemented: " ~ __traits(identifier, method));
				}
				
				
				/*
				executeInstanceMethodReturnEnumArgsNone
				executeInstanceMethodReturnIntArgsNone
				executeInstanceMethodReturnReferenceArgsInt
				executeInstanceMethodReturnNoneArgsStructFloat
				executeInstanceMethodReturnNoneArgsStructStructFloat
				executeInstanceMethodReturnNoneArgsStructStructFloatReference
				executeInstanceMethodReturnNoneArgsReferenceStructStructFloatBoolean
				executeInstanceMethodReturnNoneArgsReferenceFloat
				executeInstanceMethodReturnNoneArgsFloatFloatFloatFloatFloat
				executeInstanceMethodReturnReferenceArgsString
				*/
			}
		}
	}
	
	import std.traits;
	
	this(int reference)
	{
		super(reference);
	}

	static foreach(memberName; __traits(derivedMembers, typeof(this))) {
		static if(isSomeFunction!(__traits(getMember, typeof(this), memberName)))
		{
			static foreach(oi, overload; __traits(getOverloads, typeof(this), memberName))
			{
				static if(!hasUDA!(overload, skip) && memberName != "__ctor")
				{
					mixin DelphiImportImpl!(typeof(this), overload, oi);
				}
				
				
			}
		}
	}
}





void executeInstanceMethodReturnNoneArgsNone(int reference, string name)
{
	alias extern(Windows) void function(int, char*) FN;
	auto pChar = toUTFz!(char*)(name);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeInstanceMethodReturnNoneArgsNone");
	auto fn = cast(FN) fp;
	fn(reference, pChar);
}

string executeInstanceMethodReturnEnumArgsNone(int reference, string name)
{
	alias extern(Windows) void function(int, char*, out BSTR) FN;
	auto pChar = toUTFz!(char*)(name);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeInstanceMethodReturnEnumArgsNone");
	auto fn = cast(FN) fp;
	
	BSTR bStr;
	fn(reference, pChar, bStr);
	
	string s = to!string(bStr[0 .. SysStringLen(bStr)]);
	SysFreeString(bStr);
	
	return s;
}

int executeInstanceMethodReturnIntArgsNone(int reference, string name)
{
	alias extern(Windows) int function(int, char*) FN;
	auto pCharName = toUTFz!(char*)(name);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeInstanceMethodReturnIntArgsNone");
	auto fn = cast(FN) fp;
	return fn(reference, pCharName);
}

bool executeInstanceMethodReturnBoolArgsNone(int reference, string name)
{
	alias extern(Windows) bool function(int, char*) FN;
	auto pCharName = toUTFz!(char*)(name);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeInstanceMethodReturnBoolArgsNone");
	auto fn = cast(FN) fp;
	return fn(reference, pCharName);
}

int executeInstanceMethodReturnIntArgsString(int reference, string name, string value)
{
	alias extern(Windows) int function(int, char*, char*) FN;
	auto pCharName = toUTFz!(char*)(name);
	auto pCharValue = toUTFz!(char*)(value);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeInstanceMethodReturnIntArgsString");
	auto fn = cast(FN) fp;
	return fn(reference, pCharName, pCharValue);
}

void executeInstanceMethodReturnNoneArgsString(int reference, string name, string value)
{
	alias extern(Windows) void function(int, char*, char*) FN;
	auto pCharName = toUTFz!(char*)(name);
	auto pCharValue = toUTFz!(char*)(value);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeInstanceMethodReturnNoneArgsString");
	auto fn = cast(FN) fp;
	fn(reference, pCharName, pCharValue);
}

int executeInstanceMethodReturnReferenceArgsNone(int reference, string name)
{
	alias extern(Windows) int function(int, char*) FN;
	auto pCharName = toUTFz!(char*)(name);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeInstanceMethodReturnReferenceArgsNone");
	auto fn = cast(FN) fp;
	return fn(reference, pCharName);
}

int executeInstanceMethodReturnReferenceArgsInt(int reference, string name, int i)
{
	alias extern(Windows) int function(int, char*, int) FN;
	auto pCharName = toUTFz!(char*)(name);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeInstanceMethodReturnReferenceArgsInt");
	auto fn = cast(FN) fp;
	return fn(reference, pCharName, i);
}

void executeInstanceMethodReturnNoneArgsReference(int reference, string name, int reference2)
{
	auto pCharName = toUTFz!(char*)(name);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeInstanceMethodReturnNoneArgsReference");
	alias extern(Windows) void function(int, char*, int) FN;
	auto fn = cast(FN) fp;
	fn(reference, pCharName, reference2);
}

void executeInstanceMethodReturnNoneArgsStructFloat(int reference, string name, int reference2, float f)
{
	auto pCharName = toUTFz!(char*)(name);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeInstanceMethodReturnNoneArgsStructFloat");
	alias extern(Windows) void function(int, char*, int, float) FN;
	auto fn = cast(FN) fp;
	fn(reference, pCharName, reference2, f);
}

void executeInstanceMethodReturnNoneArgsStructStructFloat(int reference, string name, int reference2, int reference3, float f)
{
	auto pCharName = toUTFz!(char*)(name);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeInstanceMethodReturnNoneArgsStructStructFloat");
	alias extern(Windows) void function(int, char*, int, int, float) FN;
	auto fn = cast(FN) fp;
	fn(reference, pCharName, reference2, reference3, f);
}

void executeInstanceMethodReturnNoneArgsStructStructFloatReference(int reference, string name, int reference2, int reference3, float f, int reference4)
{
	auto pCharName = toUTFz!(char*)(name);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeInstanceMethodReturnNoneArgsStructStructFloatReference");
	alias extern(Windows) void function(int, char*, int, int, float, int) FN;
	auto fn = cast(FN) fp;
	fn(reference, pCharName, reference2, reference3, f, reference4);
}

void executeInstanceMethodReturnNoneArgsReferenceStructStructFloatBoolean(int reference, string name, int reference2, int reference3, int reference4, float f, bool b)
{
	auto pCharName = toUTFz!(char*)(name);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeInstanceMethodReturnNoneArgsReferenceStructStructFloatBoolean");
	alias extern(Windows) void function(int, char*, int, int, int, float, bool) FN;
	auto fn = cast(FN) fp;
	fn(reference, pCharName, reference2, reference3, reference4, f, b);	
}

void executeInstanceMethodReturnNoneArgsReferenceFloat(int reference, string name, int reference2, float f)
{
	auto pCharName = toUTFz!(char*)(name);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeInstanceMethodReturnNoneArgsReferenceFloat");
	alias extern(Windows) void function(int, char*, int, float) FN;
	auto fn = cast(FN) fp;
	fn(reference, pCharName, reference2,f);
}

void executeInstanceMethodReturnNoneArgsFloatFloatFloatFloatFloat(int reference, string name, float f1, float f2, float f3, float f4, float f5)
{
	auto pCharName = toUTFz!(char*)(name);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeInstanceMethodReturnNoneArgsFloatFloatFloatFloatFloat");
	alias extern(Windows) void function(int, char*, float, float, float, float, float) FN;
	auto fn = cast(FN) fp;
	fn(reference, pCharName, f1, f2, f3, f4, f5);
}

int executeInstanceMethodReturnReferenceArgsString(int reference, string name, string value)
{
	alias extern(Windows) int function(int, char*, char*) FN;
	
	auto pCharName = toUTFz!(char*)(name);
	auto pCharValue = toUTFz!(char*)(value);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeInstanceMethodReturnReferenceArgsString");
	auto fn = cast(FN) fp;
	return fn(reference, pCharName, pCharValue);
}

string executeInstanceMethodReturnStringArgsNone(int reference, string name)
{
	alias extern(Windows) void function(int, char*, out BSTR) FN;
	
	auto pCharName = toUTFz!(char*)(name);

	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeInstanceMethodReturnStringArgsNone_Out_String");
	auto fn = cast(FN) fp;
	
	BSTR bStr;
	fn(reference, pCharName, bStr);
	string s = to!string(bStr[0 .. SysStringLen(bStr)]);
	SysFreeString(bStr);
	return s;
}

int executeInstanceMethodReturnIntArgsStringReference(int reference, string name, string value, int r)
{
	alias extern(Windows) int function(int, char*, char*, int) FN;
	auto pCharName = toUTFz!(char*)(name);
	auto pCharValue = toUTFz!(char*)(value);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeInstanceMethodReturnIntArgsStringReference");
	auto fn = cast(FN) fp;
	return fn(reference, pCharName, pCharValue, r);
}

int executeClassMethodReturnReferenceArgsString(string qualifiedName, string name, string value)
{
	alias extern(Windows) int function(char*, char*, char*) FN;
	auto pCharQualifiedName = toUTFz!(char*)(qualifiedName);
	auto pCharName = toUTFz!(char*)(name);
	auto pCharValue = toUTFz!(char*)(value);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeClassMethodReturnReferenceArgsString");
	auto fn = cast(FN) fp;
	return fn(pCharQualifiedName, pCharName, pCharValue);
}

int executeClassMethodReturnReferenceArgsNone(string qualifiedName, string name)
{
	alias extern(Windows) int function(char*, char*) FN;
	auto pCharQualifiedName = toUTFz!(char*)(qualifiedName);
	auto pCharName = toUTFz!(char*)(name);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeClassMethodReturnReferenceArgsNone");
	auto fn = cast(FN) fp;
	return fn(pCharQualifiedName, pCharName);
}

void executeClassMethodReturnNoneArgsNone(string qualifiedName, string name)
{
	alias extern(Windows) void function(char*, char*) FN;
	auto pCharQualifiedName = toUTFz!(char*)(qualifiedName);
	auto pCharName = toUTFz!(char*)(name);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeClassMethodReturnNoneArgsNone");
	auto fn = cast(FN) fp;
	fn(pCharQualifiedName, pCharName);
}

int executeClassMethodReturnReferenceArgsReference(string qualifiedName, string name, int reference)
{
	alias extern(Windows) int function(char*, char*, int) FN;
	
	auto pCharQualifiedName = toUTFz!(char*)(qualifiedName);
	auto pCharName = toUTFz!(char*)(name);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeClassMethodReturnReferenceArgsReference");
	auto fn = cast(FN) fp;
	return fn(pCharQualifiedName, pCharName, reference);
}

int executeClassMethodReturnReferenceArgsIntUInt(string qualifiedName, string name, int i1, uint i2)
{
	alias extern(Windows) int function(char*, char*, int, uint) FN;
	
	auto pCharQualifiedName = toUTFz!(char*)(qualifiedName);
	auto pCharName = toUTFz!(char*)(name);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeClassMethodReturnReferenceArgsIntUInt");
	auto fn = cast(FN) fp;
	return fn(pCharQualifiedName, pCharName, i1, i2);
}


private BSTR allocateBSTR(string s)
{
	wstring ws = s.to!wstring;
	return allocateBSTR(ws);
}

private BSTR allocateBSTR(wstring ws)
{
	return SysAllocStringLen(ws.ptr, cast(UINT) ws.length);
}

void executeClassMethodReturnNoneArgsString(string qualifiedName, string name, string value)
{
	alias extern(Windows) void function(char*, char*, BSTR) FN;
	auto pCharQualifiedName = toUTFz!(char*)(qualifiedName);
	auto pCharName = toUTFz!(char*)(name);

	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeClassMethodReturnNoneArgsString");
	auto fn = cast(FN) fp;
	
	BSTR bStr = allocateBSTR(value);
	fn(pCharQualifiedName, pCharName, bStr);
	SysFreeString(bStr);
}

string executeClassMethodReturnStringArgsStringStringString(string qualifiedName, string name, string v1, string v2, string v3)
{
	alias extern(Windows) void function(char*, char*, BSTR, BSTR, BSTR, out BSTR) FN;
	
	auto pCharQualifiedName = toUTFz!(char*)(qualifiedName);
	auto pCharName = toUTFz!(char*)(name);

	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeClassMethodReturnNoneArgsStringStringString_Out_String");
	auto fn = cast(FN) fp;

	BSTR bStr1 = allocateBSTR(v1);
	BSTR bStr2 = allocateBSTR(v2);
	BSTR bStr3 = allocateBSTR(v3);
	
	BSTR bStr;
	fn(pCharQualifiedName, pCharName, bStr1, bStr2, bStr3, bStr);
	SysFreeString(bStr1);
	SysFreeString(bStr2);
	SysFreeString(bStr3);
	
	string s = to!string(bStr[0 .. SysStringLen(bStr)]);
	SysFreeString(bStr);
	
	return s;
}

int executeClassMethodReturnReferenceArgsReferenceInt(string qualifiedName, string name, int reference, int i)
{
	alias extern(Windows) int function(char*, char*, int, int) FN;
	
	auto pCharQualifiedName = toUTFz!(char*)(qualifiedName);
	auto pCharName = toUTFz!(char*)(name);
	
	FARPROC fp = GetProcAddress(deltaLibrary.handle, "executeClassMethodReturnReferenceArgsReferenceInt");
	auto fn = cast(FN) fp;
	return fn(pCharQualifiedName, pCharName, reference, i);
}